﻿#region usings
using System;
using System.ComponentModel.Composition;
using System.Drawing;
using System.Threading;

using VVVV.PluginInterfaces.V1;
using VVVV.PluginInterfaces.V2;
using VVVV.Utils.VMath;
using VVVV.Core.Logging;

using Emgu.CV;
using Emgu.CV.Structure;
using Emgu.CV.CvEnum;
using Emgu.CV.GPU;
using System.Diagnostics;
//using ThreadState = System.Threading.ThreadState;
using System.Collections.Generic;

#endregion usings

namespace VVVV.Nodes.EmguCV
{
    //------------------------------------------

    /// <summary>
    /// clase FaceTrackingCudaFace
    /// </summary>
    class FaceTrackingCudaFace
    {
        public Vector2D Position;
        public Vector2D Scale;
    }
    class FaceTrackingCudaEye
    {
        public Vector2D Position;
        public Vector2D Scale;
    }

    //------------------------------------------
    //------------------------------------------

    /// <summary>
    /// ///////////////////////////
    /// class vvvv
    /// </summary>

    #region PluginInfo
    [PluginInfo(Name = "FaceTrackingCuda", Category = "EmguCV", Help = "Tracks faces XY and Width/Height", Tags = "")]
    #endregion PluginInfo

    public class FaceTrackingCudaNode : IPluginEvaluate, IDisposable
    {
        
        #region fields & pins
        [Input("Image", IsSingle = true)]
        ISpread<ImageRGB> FPinInImages;

        [Input("Haar Face Table", DefaultString = "", IsSingle = true, StringType = StringType.Filename)]
        IDiffSpread<string> FPath;

        [Input("Haar Eye Table", DefaultString = "haarcascade_eye.xml", IsSingle = true, StringType = StringType.Filename)]
        IDiffSpread<string> FPathEye;

        [Input("Enabled", DefaultValue = 1)]
        ISpread<bool> FEnabled;
        
        [Output("Position")]
        ISpread<Vector2D> FPinOutPositionXY;

        [Output("Scale")]
        ISpread<Vector2D> FPinOutScaleXY;
        
        /*
		[Output("Position")] 
		ISpread<ISpread<Vector2D>> FPinOutPositionXY;

		[Output("Scale")] 
		ISpread<ISpread<Vector2D>> FPinOutScaleXY;
        */
        [Output("Status")]
        ISpread<string> FStatus;

        [Import]
        ILogger FLogger;

        HaarCascade FHaarCascade;

       // Dictionary<int, FaceTrackingCudaFace> FFaceTrackingCudaFace_Dic = new Dictionary<int, FaceTrackingCudaFace>();
        List<FaceTrackingCudaFace> Faces = new List<FaceTrackingCudaFace>();
        List<FaceTrackingCudaEye> Eyes = new List<FaceTrackingCudaEye>();
        bool IsRunning;

        Stopwatch watch;

        readonly Vector2D CMinimumSourceXY = new Vector2D(0, 0);
        readonly Vector2D CMinimumDestXY = new Vector2D(-1, 1);
        readonly Vector2D CMaximumDestXY = new Vector2D(1, -1);

        #endregion fields & pins

        [ImportingConstructor]
        public FaceTrackingCudaNode(IPluginHost host)
        {

        }

        public void Dispose()
        {
            if (IsRunning)
            {
                IsRunning = false;
            }

            // foreach (KeyValuePair<int, FaceTrackingCudaInstance> tracker in FFaceTrackingCudaInstance_Dic)
            //tracker.Value.Close();
        }

        //called when data for any output pin is requested
        public void Evaluate(int SpreadMax)
        {
            if (FPath.IsChanged)
            {
                if (FPath.SliceCount > 0)
                {
                    try
                    {
                        FHaarCascade = new HaarCascade(FPath[0]);
                    }
                    catch
                    {
                        for (int i = 0; i < FStatus.SliceCount; i++)
                            FStatus[i] = "Loading cascade xml failed";
                    }
                }
            }

            if (FHaarCascade == null)
            {
                FStatus.SliceCount = 1;
                FStatus[0] = "Please load haar cascade xml";
                return;
            }

            if (FPinInImages != null && FHaarCascade != null && FPath[0] != "")
            {
                try{
                    //FLogger.Log(LogType.Debug, "hola");
                    IsRunning = true;
                    UpdatealoOne();
                }
                catch (Exception e){
                    FLogger.Log(LogType.Error, e.Message);
                    FPath[0] = "";

                }
                
            }

            //UpdateTrackers();
            //OutputFaces();
        }
        [STAThread]
        void UpdatealoOne()
        {
            if (IsRunning)
            {
                if (FPinInImages[0].FrameChanged)
                {
                    //FLogger.Log(LogType.Warning, "cambiando");
                    Faces.Clear();
                  
                    using (GpuCascadeClassifier face = new GpuCascadeClassifier(FPath[0]))
                   // using (GpuCascadeClassifier eye = new GpuCascadeClassifier(FPathEye[0]))
                    {
                        
                        watch = Stopwatch.StartNew();
                        
                        using (GpuImage<Bgr, Byte> gpuImage = new GpuImage<Bgr, byte>(FPinInImages[FPinInImages.SliceCount-1].Img))
                        using (GpuImage<Gray, Byte> gpuGray = gpuImage.Convert<Gray, Byte>())
                        {
                            Rectangle[] faceRegion = face.DetectMultiScale(gpuGray, 1.1, 10, Size.Empty);
                            
                            foreach (Rectangle f in faceRegion)
                            {
                                //draw the face detected in the 0th (gray) channel with blue color
                                //FSource.Draw(f, new Bgr(Color.Blue), 2);

                                FaceTrackingCudaFace careto = new FaceTrackingCudaFace();

                                var faceVector = new Vector2D(f.X + f.Width / 2, f.Y + f.Height / 2);

                                Vector2D CMaximumSourceXY = new Vector2D(gpuGray.Size.Width, gpuGray.Size.Height);

                                careto.Position = VMath.Map(faceVector, CMinimumSourceXY, CMaximumSourceXY, CMinimumDestXY, CMaximumDestXY, TMapMode.Float);
                                careto.Scale = VMath.Map(new Vector2D(f.Width, f.Height), CMinimumSourceXY.x, CMaximumSourceXY.x, 0, 2, TMapMode.Float);

                                Faces.Add(careto);
                                
                                /*
                                using (GpuImage<Gray, Byte> faceImg = gpuGray.GetSubRect(f))
                                {
                                    //For some reason a clone is required.
                                    //Might be a bug of GpuCascadeClassifier in opencv
                                    using (GpuImage<Gray, Byte> clone = faceImg.Clone())
                                    {
                                        Rectangle[] eyeRegion = eye.DetectMultiScale(clone, 1.1, 10, Size.Empty);

                                        foreach (Rectangle e in eyeRegion)
                                        {
                                            Rectangle eyeRect = e;
                                            eyeRect.Offset(f.X, f.Y);
                                            
                                        }
                                    }
                                }*/
                            }
                            watch.Stop();
                        }
                    }

                    FLogger.Log(LogType.Debug, "número de caras --> "+Faces.Count.ToString());
                    //float seg = watch.ElapsedMilliseconds / 1000;
                    FLogger.Log(LogType.Debug, "velocidad de reconocimiento --> " + watch.ElapsedMilliseconds.ToString() + " milisegundos");

                    FPinOutPositionXY.SliceCount = Faces.Count;
                    FPinOutScaleXY.SliceCount = Faces.Count;

                    if (Faces.Count > 0)
                    {
                        for (int i = 0; i < Faces.Count; i++)
                        {
                            FPinOutPositionXY[i] = Faces[i].Position;
                            FPinOutScaleXY[i] = Faces[i].Scale;
                        }
                    }
                    else {

                        FPinOutPositionXY[0] = new Vector2D(0, 0);
                        FPinOutScaleXY[0] = new Vector2D(0, 0);
                    }

                }
            }
        }

    }
}
